@@ -21,14 +21,15 @@
  *
  */
 
-#include "bscconfig.h"
-#include <gtp.h>
+#include "config.h"
+#include <osmocom/gtp/gtp.h>
 
 #include <osmocom/core/rate_ctr.h>
 #include <osmocom/core/tdef.h>
 #include <osmocom/gprs/gprs_msgb.h>
 
 #include <osmocom/ranap/ranap_common.h>
+#include <osmocom/ranap/iu_helpers.h>
 
 #include <osmocom/sgsn/gprs_gmm.h>
 #include <osmocom/sgsn/gprs_sm.h>
@@ -37,6 +38,10 @@
 #include <osmocom/sgsn/gprs_ranap.h>
 #include <osmocom/sgsn/gprs_gmm_attach.h>
 #include <osmocom/sgsn/gprs_mm_state_iu_fsm.h>
+#include <osmocom/sgsn/gtp_ggsn.h>
+#include <osmocom/sgsn/gtp.h>
+#include <osmocom/sgsn/pdpctx.h>
+#include <osmocom/sgsn/mmctx.h>
 
 /* Send RAB activation requests for all PDP contexts */
 void activate_pdp_rabs(struct sgsn_mm_ctx *ctx)
@@ -56,36 +61,44 @@ static int sgsn_ranap_rab_ass_resp(struct sgsn_mm_ctx *ctx, RANAP_RAB_SetupOrMod
 	bool require_pdp_update = false;
 	struct sgsn_pdp_ctx *pdp = NULL;
 	RANAP_RAB_SetupOrModifiedItem_t *item = &setup_ies->raB_SetupOrModifiedItem;
+	int rc;
 
 	rab_id = item->rAB_ID.buf[0];
 
 	pdp = sgsn_pdp_ctx_by_nsapi(ctx, rab_id);
 	if (!pdp) {
 		LOGP(DRANAP, LOGL_ERROR, "RAB Assignment Response for unknown RAB/NSAPI=%u\n", rab_id);
+		sgsn_ranap_iu_release_free(ctx, NULL);
 		return -1;
 	}
 
 	if (item->transportLayerAddress) {
+		struct osmo_sockaddr addr;
 		LOGPC(DRANAP, LOGL_INFO, " Setup: (%u/%s)", rab_id, osmo_hexdump(item->transportLayerAddress->buf,
 								     item->transportLayerAddress->size));
-		switch (item->transportLayerAddress->size) {
-		case 7:
-			/* It must be IPv4 inside a X213 NSAP */
-			memcpy(pdp->lib->gsnlu.v, &item->transportLayerAddress->buf[3], 4);
+		rc = ranap_transp_layer_addr_decode2(&addr, NULL, item->transportLayerAddress);
+		if (rc < 0) {
+			LOGP(DRANAP, LOGL_ERROR,
+			     "RAB Assignment Resp: Unknown Transport Layer Address (size %u): %s\n",
+			     item->transportLayerAddress->size,
+			     osmo_hexdump(item->transportLayerAddress->buf, item->transportLayerAddress->size));
+			goto ret_error;
+		}
+
+		switch (addr.u.sa.sa_family) {
+		case AF_INET:
+			memcpy(pdp->lib->gsnlu.v, (uint8_t *)&addr.u.sin.sin_addr.s_addr, 4);
 			break;
-		case 4:
-			/* It must be a raw IPv4 address */
-			memcpy(pdp->lib->gsnlu.v, item->transportLayerAddress->buf, 4);
-			break;
-		case 16:
-			/* TODO: It must be a raw IPv6 address */
-		case 19:
-			/* TODO: It must be IPv6 inside a X213 NSAP */
+		case AF_INET6:
+			/* TODO: Support IPv6 address */
+			LOGP(DRANAP, LOGL_ERROR,
+			     "RAB Assignment Resp: IPv6 transport layer address not supported!\n");
+			goto ret_error;
 		default:
-			LOGP(DRANAP, LOGL_ERROR, "RAB Assignment Resp: Unknown "
-				"transport layer address size %u\n",
-				item->transportLayerAddress->size);
-			return -1;
+			LOGP(DRANAP, LOGL_ERROR,
+			     "RAB Assignment Resp: Unexpected transport layer address size %u\n",
+			     item->transportLayerAddress->size);
+			goto ret_error;
 		}
 		require_pdp_update = true;
 	}
@@ -99,6 +112,8 @@ static int sgsn_ranap_rab_ass_resp(struct sgsn_mm_ctx *ctx, RANAP_RAB_SetupOrMod
 		LOGP(DRANAP, LOGL_DEBUG, "Updating TEID on RNC side from 0x%08x to 0x%08x\n",
 			pdp->lib->teid_own, tei);
 		pdp->lib->teid_own = tei;
+		pdp->lib->dir_tun_flags.l = 1;
+		pdp->lib->dir_tun_flags.v[0] = 0x01; /* Set DTI flag in Direct Tunnel Flags */
 		require_pdp_update = true;
 	}
 
@@ -111,6 +126,15 @@ static int sgsn_ranap_rab_ass_resp(struct sgsn_mm_ctx *ctx, RANAP_RAB_SetupOrMod
 	}
 	return 0;
 
+ret_error:
+	if (pdp->state != PDP_STATE_CR_CONF) {
+		gsm48_tx_gsm_act_pdp_rej(ctx, pdp->ti, GSM_CAUSE_NET_FAIL,
+					 0, NULL);
+		sgsn_delete_pdp_ctx(pdp);
+	} else {
+		gsm48_tx_gsm_deact_pdp_req(pdp, GSM_CAUSE_NET_FAIL, true);
+	}
+	return -1;
 }
 
 int sgsn_ranap_iu_event(struct ranap_ue_conn_ctx *ctx, enum ranap_iu_event_type type, void *data)
@@ -120,7 +144,8 @@ int sgsn_ranap_iu_event(struct ranap_ue_conn_ctx *ctx, enum ranap_iu_event_type
 
 	mm = sgsn_mm_ctx_by_ue_ctx(ctx);
 	if (!mm) {
-		LOGIUP(ctx, LOGL_NOTICE, "Cannot find mm ctx for IU event %d\n", type);
+		LOGIUP(ctx, LOGL_NOTICE, "Cannot find mm ctx for IU event %s\n",
+		       ranap_iu_event_type_str(type));
 		ranap_iu_free_ue(ctx);
 		return rc;
 	}
@@ -133,10 +158,9 @@ int sgsn_ranap_iu_event(struct ranap_ue_conn_ctx *ctx, enum ranap_iu_event_type
 		/* fall thru */
 	case RANAP_IU_EVENT_LINK_INVALIDATED:
 		/* Clean up ranap_ue_conn_ctx here */
-		LOGMMCTXP(LOGL_INFO, mm, "IU release for imsi %s\n", mm->imsi);
-		if (mm->iu.mm_state_fsm->state == ST_PMM_CONNECTED)
-			osmo_fsm_inst_dispatch(mm->iu.mm_state_fsm, E_PMM_PS_CONN_RELEASE, NULL);
-		else
+		LOGMMCTXP(LOGL_INFO, mm, "IU release (cause=%s)\n", ranap_iu_event_type_str(type));
+		rc = osmo_fsm_inst_dispatch(mm->iu.mm_state_fsm, E_PMM_PS_CONN_RELEASE, NULL);
+		if (rc < 0)
 			sgsn_ranap_iu_free(mm);
 
 		/* TODO: move this into FSM */
@@ -145,6 +169,11 @@ int sgsn_ranap_iu_event(struct ranap_ue_conn_ctx *ctx, enum ranap_iu_event_type
 		rc = 0;
 		break;
 	case RANAP_IU_EVENT_SECURITY_MODE_COMPLETE:
+		/* FIXME: verify that a permitted UEA level was chosen. Compare how osmo-msc does it in
+		 * msc_a_ran_dec_from_msc_i(), case RAN_MSG_CIPHER_MODE_COMPLETE.
+		 * We should dissolve iu_client.c, it was a design mistake when first implementing Iu support. osmo-msc
+		 * has moved away from it a long time ago.
+		 */
 		/* Continue authentication here */
 		mm->iu.ue_ctx->integrity_active = 1;
 		ranap_iu_tx_common_id(mm->iu.ue_ctx, mm->imsi);
@@ -157,7 +186,7 @@ int sgsn_ranap_iu_event(struct ranap_ue_conn_ctx *ctx, enum ranap_iu_event_type
 		rc = 0;
 		break;
 	default:
-		LOGMMCTXP(LOGL_NOTICE, mm, "Unknown event received: %i\n", type);
+		LOGMMCTXP(LOGL_NOTICE, mm, "Unknown event received: %d\n", type);
 		rc = -1;
 		break;
 	}
@@ -232,7 +261,7 @@ int gsm0408_gprs_rcvmsg_iu(struct msgb *msg, struct gprs_ra_id *ra_id,
 
 	mmctx = sgsn_mm_ctx_by_ue_ctx(MSG_IU_UE_CTX(msg));
 	if (mmctx) {
-		rate_ctr_inc(&mmctx->ctrg->ctr[GMM_CTR_PKTS_SIG_IN]);
+		rate_ctr_inc(rate_ctr_group_get_ctr(mmctx->ctrg, GMM_CTR_PKTS_SIG_IN));
 		if (ra_id)
 			memcpy(&mmctx->ra, ra_id, sizeof(mmctx->ra));
 	}
